/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: FWObject.cc,v 1.50 2001/06/24 07:25:08 vkurland Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "Tools.hh"
#include "FWObject.hh"
#include "FWObjectDialog.hh"
#include "FWObjectDatabase.hh"
#include "FWObjectReference.hh"
#include "DialogPlugin.hh"
#include "helpers.hh"
#include "main_window.hh"

#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <fcntl.h>

#include <algorithm>

#include "XMLTools.hh"

#include <algorithm>

const char *FWObject::TYPENAME={"UNDEF"};

string FWObject::NOT_FOUND="";

void FWObject::fromXML(xmlNodePtr root)
{
    assert(root!=NULL);

    const char *n;

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("comment")));
    if(n)  setComment(n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("name")));
    if(n)  setName(n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("id")));
    if(n)  setId(n);


    /******************************************/

    ref_counter=0;

    for(xmlNodePtr cur=root->xmlChildrenNode; cur; cur=cur->next) 
    {
        if(cur && !xmlIsBlankNode(cur))  
        {
            FWObject *o=FWObjectDatabase::createFromXML(cur);
            if(o!=NULL) 
            {
		add(o);
		o->fromXML(cur);
	    }
        }
    }
    
    setDirty(false);
}


xmlNodePtr FWObject::toXML(xmlNodePtr parent)
{
    xmlNodePtr me = xmlNewChild(parent, NULL, xml_name.empty()?STRTOXMLCAST(getTypeName()):STRTOXMLCAST(xml_name), NULL);

    FWObject *o;
    
    map<string, string>::iterator i;
    for(i=data.begin(); i!=data.end(); ++i) 
    {
        string name  = (*i).first;
        string value = (*i).second;

        xmlAttrPtr pr = xmlNewProp(me, 
                                   STRTOXMLCAST(name) , 
                                   STRTOXMLCAST(value));
        if(name=="id")
        {
            xmlAddID(NULL, parent->doc, STRTOXMLCAST(value), pr);
        } else if(name=="ref") 
        {
            xmlAddRef(NULL, parent->doc, STRTOXMLCAST(value), pr);
        }
    }
    
    vector<FWObject*>::iterator j;
    for(j=begin(); j!=end(); ++j) 
    {
        if((o=(*j))!=NULL) o->toXML(me);
    }
    
    return me;
}


FWObject::FWObject()
{
    ref_counter = 0;
    parent      = NULL;

    // When object created we assign it unique Id
    setId(FWObjectDatabase::generateUniqueId());

    setDirty(false);
}

FWObject::FWObject(FWObject &c)
{
    *this=c;
}


FWObject::FWObject(FWObject *par)
{
    setName("New object");
    
    ref_counter = 0    ;
    parent      = par  ;
    
    // When object created we assign it unique Id
    setId(FWObjectDatabase::generateUniqueId());

    setDirty(false);
}


FWObject::~FWObject() 
{
    clearChildren();
    data.clear();
}

FWObject* FWObject::getParent()
{
    return parent;
}

void FWObject::setParent(FWObject *p)
{
    parent=p;
}


/**
 *  Rarely used feature: we can change the name of XML element
 *  represented by the class derived from FWObject if we want to. For
 *  example, used in all "Any"  objects so we did not have to create extra
 *  classes
 * 
 */
void FWObject::setXMLName(const string &n)
{
    xml_name = n;
}


bool FWObject::isSystem()
{
    return Resources::global_res->getObjResourceBool(this,"system");
}



FWObject* FWObject::_find(const string& name)   // find child by name
{
    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
	if (  (o=(*m))!=NULL && 
	      o->getName()==name )
	    return(o);
    }
    return(NULL);
}

bool      FWObject::_child(FWObject *obj)       // verifies if obj is child
{
    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
	if (  (o=(*m))!=NULL && o==obj) return(true);
    }
    return(false);
}

/**
 *  This method copies content of object 'x' to object 'this', including
 *  ID. Resultant object is identical copy of 'x'.
 *  This is mostly used for creating scratchpad copies
 */
FWObject& FWObject::operator=(const FWObject &x)
{
    clearChildren();
    data.clear();

    ref_counter = 0;
    parent      = x.parent;

    xml_name    = x.xml_name;
    
    data=x.data;


    vector<FWObject*>::const_iterator m;
    for(m=x.begin(); m!=x.end(); ++m) 
    {
        FWObject *o;
	if((o=(*m))!=NULL) 
        {
            FWObject *o1 = FWObjectDatabase::create( o->getTypeName() );
	    *o1=*o;
	    add(o1);
	}
    }
    
    setDirty(true);

    return *this;
}

/**
 *   this method copies content of object 'x' in the object 'this'
 *   preserving object's 'this' ID
 *
 *   Thus, what we get in the end is still unique object, but its content
 *   is the same as that of the object 'x'
 */
FWObject& FWObject::duplicate(FWObject &x)
{
    string id=getId();
    
    clearChildren();
    data.clear();

    data=x.data;

    if (id!="")   // some objects do not have ID per DTD (e.g. Src, Dst, etc.)
	setId(id);

    FWObject *o, *o1;

    vector<FWObject*>::iterator m;
    for (m=x.begin(); m!=x.end(); ++m) 
    {
	if((o=(*m))!=NULL) 
        {
	    o1=FWObjectDatabase::create( o->getTypeName() );
	    o1->duplicate(*o);
	    add(o1);
	}
    }
    
    setDirty(true);

    return *this;
}

const string &FWObject::getName() const 
{ 
    return getStr("name"); 
}

void FWObject::setName(const string &n)   
{
    setStr("name",n);
    setDirty(true);
}


const string &FWObject::getComment() const
{ 
    return getStr("comment"); 
}

void FWObject::setComment(const string &c)
{
    setStr("comment",c);
    setDirty(true);
}

const string &FWObject::getId() const
{ 
    return getStr("id");
}

void FWObject::setId(const string &c)
{
    setStr("id",c);
    setDirty(true);
}

bool FWObject::exists(const string &name) const 
{
    return data.count(name)!=0; 
}

/*************************************************************************
 *      Str
 */
const string &FWObject::getStr(const string &name) const
{
    if(exists(name)) 
    {
        map<string,string>::const_iterator i=data.find(name);
        return (*i).second;
    } else
    {
        return NOT_FOUND;
    }
}

void FWObject::remStr(const string &name)
{
    if(exists(name)) 
    {
	map<string, string>::iterator m=data.find(name);
	if(m != data.end()) 
        {
	    data.erase(m);
	    setDirty(true);
	}
    }
}

void FWObject::setStr(const string &name, const string &val)
{
    data[name]=val;
    setDirty(true);
}

/*************************************************************************
 *       Int
 */

int FWObject::getInt(const string &name) const
{
    string s=getStr(name);
    if(s!="") 
        return( atol(s.c_str()) );
    else   
        return(-1);
}

void FWObject::setInt(const string &name, int val)
{
    gchar str[128];
    sprintf(str,"%d",val);
    setStr(name, str);
    setDirty(true);
}

/*************************************************************************
 *       Bool
 */
bool FWObject::getBool(const string &name) const
{
    if(exists(name)) 
    {
	return(getStr(name)=="1" || 
               strcasecmp(getStr(name).c_str() , "true")==0);
    } else
        return false;
}

void FWObject::setBool(const string &name, bool val)
{
    setStr(name, (val)?"True":"False");
    setDirty(true);
}

void FWObject::setBool(const string &name, const string &val)
{
    if(!name.empty())
	setBool(name,
		(val=="1" || strcasecmp(val.c_str(),"true")==0)); 
}

/*
************************************************************************
 */
string  FWObject::getXPMIconFileName(const string& icon_sfx)
{ 
    string res;

    res=
	Resources::global_res->getResourceStr("/FWBuilderResources/Paths/Icndir");
    res += "/";
    res += Resources::global_res->getObjResourceStr(this,"icon"+icon_sfx);

    return res;
}

void FWObject::Show()
{
    setBool("read",true);
}

void FWObject::Hide()
{
    setBool("read",false);
}

DialogPlugin* FWObject::CreateDialog()
{
    return ( new FWObjectDialog(this) );
}

/*
 *  disable menu items which make sence only for specific objects
 */
void   FWObject::updateMainMenu()
{
    Gtk::MenuBar *main_menubar=main_w->getMainMenuBar();

    Gtk::Widget *itm=find_widget("policy",main_menubar);
    if (itm) itm->set_sensitive(false);

    itm=find_widget("duplicate",main_menubar);
    if (itm) itm->set_sensitive( ! isSystem() );

    itm=find_widget("copy",main_menubar);
    if (itm) itm->set_sensitive( ! isSystem() );

    itm=find_widget("cut",main_menubar);
    if (itm) itm->set_sensitive( ! isSystem() );

    itm=find_widget("paste",main_menubar);
    if (itm) itm->set_sensitive( true );

    itm=find_widget("sort",main_menubar);
    if (itm) itm->set_sensitive( isSystem() );
}

void   FWObject::dump(bool recursive,bool brief,int offset)
{
    std::ofstream out( STDERR_FILENO );
    dump(out,recursive,brief,offset);
}

void   FWObject::dump(std::ofstream &f,bool recursive,bool brief,int offset)
{
    FWObject *o;
    string    n;

    if (brief) {
	f << string(offset,' ');
	f << " Obj=" << this;
	f << " ID="  << getId();
	f << " Name=" << getName();
	f << " Type=" << getTypeName();

	if (FWReference::cast(this)!=0) {
	    f << " Ref=" << FWReference::cast(this)->getPointer();
	}

	f << endl;

	if (recursive) {
	    vector<FWObject*>::iterator m;
	    for (m=begin(); m!=end(); ++m) {
		if (  (o=(*m))!=NULL)  o->dump(f,false,brief,offset+2);
	    }
	}
    } else {

	f << string(offset,' ') << string(16,'-') << endl;
	f << string(offset,' ') << "Obj:    " << this << endl;
	f << string(offset,' ') << "ID:     " << getId() << endl;
	f << string(offset,' ') << "Name:   " << getName() << endl;
	f << string(offset,' ') << "Ref.ctr:" << ref_counter << endl;
	f << string(offset,' ') << "Type:   " << getTypeName() << endl;
//    f << string(offset,' ') << "Path:   " << getPath() << endl;
	n=(getParent()!=NULL)?getParent()->getName():"";
	f << string(offset,' ') << "Parent: " << getParent()
	  << "  name=" << n << endl;
	f << string(offset,' ') << "system: " << isSystem() << endl;

	map<string, string>::iterator d;
	for (d=data.begin(); d!=data.end(); ++d) {
	    if((*d).first=="name") 
                continue;
	    f << string(offset,' ');
	    f << (*d).first << ": " << (*d).second << endl;
	}
	if (recursive) {
	    vector<FWObject*>::iterator m;
	    for (m=begin(); m!=end(); ++m) {
		if (  (o=(*m))!=NULL)  o->dump(f,recursive,brief,offset+2);
	    }
	}
    }

}

void   FWObject::_adopt(FWObject *obj)
{
    if (obj==NULL) return;

    obj->ref();
    obj->setParent(this);
}

void FWObject::addAt(const string& where_id, FWObject *obj)
{
    FWObject *p=getById( where_id , true );
    assert (p!=NULL);
    p->add(obj);
}

void   FWObject::add(FWObject *obj)
{
    if (obj==NULL) return;

    if ( validateChild(obj) ) {

	push_back(obj);
	_adopt(obj);
	setDirty(true);
    }
}

FWReference* FWObject::createRef( FWObject* obj )
{
    return new FWObjectReference(obj);
}


void FWObject::addRef(FWObject *obj)
{
    if (obj==NULL) return;

    if ( validateChild(obj) ) {
	    
	FWReference *oref=createRef(obj);
	obj->ref();

	push_back(oref);
	_adopt(oref);
	setDirty(true);
    }
}

void FWObject::insert_before(FWObject *o1,FWObject *obj)
{
    if (obj==NULL) return;

    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
	if ( (o=(*m))!=NULL && o==o1) {
	    insert(m,obj);
	    _adopt(obj);
	    setDirty(true);
	    return;
	}
    }
}

void FWObject::insert_after(FWObject *o1,FWObject *obj)
{
    if (obj==NULL) return;

    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
	if ( (o=(*m))!=NULL && o==o1) {
	    insert(++m,obj);
	    _adopt(obj);
	    setDirty(true);
	    return;
	}
    }
}

void FWObject::swapObjects(FWObject *o1,FWObject *o2)
{
    FWObject *o;
    vector<FWObject*>::iterator m, m1, m2;
    for (m=begin(); m!=end(); ++m) {
	if ( (o=(*m))!=NULL ) {
	    if ( o==o1 ) m1=m;
	    if ( o==o2 ) m2=m;
	}
    }
    if ( (*m1)!=NULL && (*m2)!=NULL ) {
	*m1=o2;
	*m2=o1;
    }
    setDirty(true);
}

void FWObject::_remove(FWObject *obj)
{
    if (obj==NULL) return;

    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
	if (  (o=(*m))!=NULL && o==obj) {
	    erase(m);
	    obj->unref();
	    setDirty(true);
	    return;
	}
    }
}


void FWObject::remove(FWObject *obj)
{
    if (obj==NULL) return;

    _remove(obj);
    if(obj->ref_counter==0) 
        delete obj;
}

void FWObject::_removeAllInstances(FWObject *obj)
{
    if (obj==NULL) return;

    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
        if (  (o=(*m))!=NULL ) o->_removeAllInstances(obj);
    }

    _remove(obj);

    removeRef(obj);
}

void FWObject::removeAllInstances(FWObject *obj)
{
    if (obj==NULL) return;
    _removeAllInstances(obj);
    if (obj->ref_counter==0) delete obj;
}

bool FWObject::validateChild(FWObject *obj)
{ 
    return true;

/*
 *  Check if object "this" is a descendant of object "obj" to avoid loops
 *
 *  check disabled for now since we need to be able to add firewall to its
 *  own policy
 */
    FWObject *p;
    p=this;
    do {
	if (p==obj) return false;
    } while ((p=p->getParent())!=NULL);
    return true;
}


void FWObject::removeRef(FWObject *obj)
{
    FWObject *o;

    if (obj==NULL) return;

    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) {
        if ( (o=(*m))!=NULL && FWReference::cast(o)!=NULL) {
	    FWReference *oref=FWReference::cast(o);
	    if (oref!=NULL && oref->getPointer()==obj) {
// do not delete object even if this reference was the last one (?)
		obj->unref();  
// remove will delete o  'cause reference counter goes to zero
		FWObject::remove(o); 
		return;
	    }
	}
    }
}

void FWObject::clearChildren()
{
    FWObject *o;
    
    vector<FWObject*>::iterator m;
    for(m=begin(); m!=end(); ++m) 
    {
	if((o=(*m))!=NULL) 
        {
            o->unref();
            if(o->ref_counter==0) 
                delete o;
	}
    }
    clear();
    setDirty(true);
}

void FWObject::clearChildren(const string &type_name)
{
    vector<FWObject*>::iterator m;
 loop:
    for(m=begin(); m!=end(); ++m) 
    {
        FWObject *o=(*m);
        if (o!=NULL && o->getTypeName()==type_name) 
        {
	    remove(o); 
            goto loop;
        }
    }
}

int  FWObject::getChildrenCount()
{
    return(size());
}

/*
void FWObject::resolveReferences()
{
    FWObject *o;
    vector<FWObject*>::iterator m;
    for (m=begin(); m!=end(); ++m) 
    {
        if ( (o=(*m))!=NULL )  o->resolveReferences();
    }
}
*/

/**
 *   Returns root of the objects tree this object is part of
 */
FWObject* FWObject::getRoot()
{
    FWObject *p=this;

    while (p->getParent()!=NULL) p=p->getParent();
    return p;
}


/**
 * Returns object with given ID. If recursive is 'false'
 * search is restricted to direct children of given object
 * and itslef. If it is tru it is proceed to all chidren of all
 * objects starting from current one.
 */
FWObject* FWObject::getById  (const string &id, bool recursive)
{
    if(id==getId())
        return this;
    
    vector<FWObject*>::iterator j;
    for(j=begin(); j!=end(); ++j)     
    {
        FWObject *o;
        if((o=(*j))!=NULL)   
        {
	    string oid=o->getId();
	    if(id==oid) 
                return o;
            
            if(recursive)  
            {
                o=o->getById(id, true);
                if(o) 
                    return o;
            }
        }
    }
    return NULL; // not found
}


FWObject* FWObject::getFirstByType(const string &type_name)
{
    vector<FWObject*>::iterator j;
    for(j=begin(); j!=end(); ++j)     {
        FWObject *o;
        if((o=(*j))!=NULL)   {
            if(type_name == o->getTypeName())
                return o;
        }
    }
    return NULL;
}

vector<FWObject*> FWObject::getByType(const string &type_name)
{
    vector<FWObject*> res;

    vector<FWObject*>::iterator j;
    for(j=begin(); j!=end(); ++j)     {
        FWObject *o;
        if((o=(*j))!=NULL)  {
            if(type_name == o->getTypeName())
                res.push_back(o);
        }
    }
    return res; 
}

void  FWObject::setDirty(bool f,bool recursive)
{
    dirty=f;
    if(recursive) 
    {
	FWObject *o;
	vector<FWObject*>::iterator m;
	for (m=begin(); m!=end(); ++m) 
	    if ( (o=(*m))!=NULL )  o->setDirty(f, TRUE);
    }
}

bool FWObject::isDirty(bool recursive)
{
    if(dirty)
        return TRUE;

    if(recursive) 
    {
	FWObject *o;
	vector<FWObject*>::iterator m;
	for(m=begin(); m!=end(); ++m) 
	    if((o=(*m))!=NULL)  
		if(o->isDirty(recursive))
                    return TRUE;
    }
    return FALSE;
}

bool FWObject::GUISortOrder(const FWObject *a, const FWObject *b)
{
    return a->getName() < b->getName();
}

class sort_order_func_adaptor 
{
    private:
    
    FWObject *o;
    
    public:
    
    explicit sort_order_func_adaptor(FWObject *p) { o=p; }
    
    bool operator()(const FWObject *a, const FWObject *b) 
    { 
        return o->GUISortOrder(a,b);
    }
};

void FWObject::sortChildren()
{
    std::sort(begin(), end(), sort_order_func_adaptor(this));
    setDirty(true);
}
