/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: FWObject.hh,v 1.31 2001/05/25 02:14:41 lord Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef  __FWOBJECT_HH
#define  __FWOBJECT_HH

#include <string>
#include <vector>
#include <map>

#include <gtk--.h>

#include <parser.h>
#include <tree.h>

#include <fstream>
#include <cstdlib>


class DialogPlugin;

class FWObject;
class FWReference;

#define DECLARE_FWOBJECT_SUBTYPE(name) \
   static    const char *TYPENAME; \
   virtual   string getTypeName() const { return TYPENAME; } \
   static bool isA(const FWObject *o) { return o && o->getTypeName()==TYPENAME; }  \
   static name* cast(FWObject *o) { return dynamic_cast<name*>(o); }

/**
 * This class represents base class for all
 * objects we are dealing with.
 *
 * FWObject can have children, that is other objects of the same class
 * or derived classes which are included in this one
 */
class FWObject : public vector<FWObject*>
{

    friend class FWObjectDialog;

private:

    int            ref_counter;
    bool           dirty;
    FWObject      *parent;
    static string NOT_FOUND;

protected:

    string         xml_name;

    map<string, string> data;
    
    void      _remove(FWObject *obj);
    void      _removeAllInstances(FWObject *obj);
    void      _adopt(FWObject *obj);   // increments reference
    
    FWObject* _find(const string& name);   // find child by name
    bool      _child(FWObject *obj);       // verifies if obj is child
    
    FWObject();
    FWObject(FWObject &copy);
    FWObject(FWObject *parent);

public:

    DECLARE_FWOBJECT_SUBTYPE(FWObject);

    virtual void       fromXML    (xmlNodePtr xml_parent_node);
    virtual xmlNodePtr toXML      (xmlNodePtr xml_parent_node);
            void       setXMLName (const string &);

    virtual  ~FWObject();

    int ref()   { ++ref_counter; return(ref_counter); }
    int unref() { --ref_counter; return(ref_counter); }

    virtual FWObject& operator=(const FWObject &);

    FWObject& duplicate(FWObject &obj);
    
    string    getXPMIconFileName(const string& icon_sfx);

    void      Show();
    void      Hide();

    FWObject* getParent();
    void      setParent(FWObject *p);

    bool    exists(const string &name) const;

    void    remStr(const string &name);

    const string &getStr(const string& name) const;
    void    setStr(const string &name, const string &val);

    int     getInt(const string &name) const;
    void    setInt(const string &name, int    val);

    bool    getBool(const string &name) const;
    void    setBool(const string &name, bool val);
    void    setBool(const string &name, const string &val);

    const string &getName() const;
    void          setName(const string& n);
    
    const string &getComment() const;
    void          setComment(const string& c);

    const string  &getId() const;
    void           setId(const string& c);

    bool    isSystem();

            void addAt(const string& where_id, FWObject *obj);
    virtual void add(FWObject *obj);
    virtual void insert_before(FWObject *o1,FWObject *obj);
    virtual void insert_after(FWObject *o1,FWObject *obj);
    virtual void swapObjects(FWObject *o1,FWObject *o2);

    virtual bool GUISortOrder(const FWObject *a, const FWObject *b);
    virtual void sortChildren();
    
    virtual void remove(FWObject *obj);
    virtual void removeAllInstances(FWObject *obj);

    virtual void addRef(FWObject *obj);
    virtual void removeRef(FWObject *obj);

    /*
     *  Create appropriate reference object for a given object.
     *  Different object require different class of reference. e.g.:
     *  for RuleElementSrc reference should be of class FWObjectReference
     *  for RuleElementSrv reference should be of class FWServiceReference
     */
    virtual FWReference* createRef( FWObject* obj );

    /*
     * verify whether given object type is approppriate as a child
     */
    virtual bool  validateChild(FWObject *o);

    void clearChildren();
    void clearChildren(const string &type_name);
    
    int  getChildrenCount();

    virtual DialogPlugin* CreateDialog();
    virtual void          updateMainMenu();

    virtual void    dump(bool recursive,bool brief,int offset=0);
    virtual void    dump(std::ofstream &f,bool recursive,bool brief,int offset=0);

    /**
     *   Returns root of the objects tree this object is part of
     */
    FWObject* getRoot();

    /**
     * Returns object with given ID. If recursive is 'false'
     * search is restricted to direct children of given object
     * and itslef. If it is tru it is proceed to all chidren of all
     * objects starting from current one.
     */
    virtual FWObject* getById  (const string &id, bool recursive=FALSE);

    /**
     * Returns list of direct children of current object
     * whose getTypeName() same as given.
     * If not found, empty list is returned.
     */
    virtual vector<FWObject*> getByType(const string &type_name);

    /**
     * Returns first of direct children of current object
     * whose getTypeName() same as given or NULL if not found.
     */
    virtual FWObject* getFirstByType(const string &type_name);

    void    setDirty(bool f,bool recursive=false);
    bool    isDirty(bool recursive);
};

#endif // _FWOBJECT_HH







