/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: FWObjectDatabase.cc,v 1.66 2001/05/25 01:45:29 lord Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <config.h>

#include "FWObject.hh"
#include "FWObjectDatabase.hh"

#include "FWOptions.hh"

#include "Rule.hh"
#include "RuleElement.hh"
#include "Policy.hh"
#include "InterfacePolicy.hh"
#include "NAT.hh"
#include "ObjectGroup.hh"
#include "ServiceGroup.hh"
#include "IntervalGroup.hh"
#include "Host.hh"
#include "Firewall.hh"
#include "Network.hh"
#include "Interface.hh"
#include "FWObjectReference.hh"
#include "FWServiceReference.hh"
#include "FWIntervalReference.hh"
#include "TCPService.hh"
#include "UDPService.hh"
#include "ICMPService.hh"
#include "IPService.hh"
#include "Interval.hh"

#include "FileSel.hh"

#include "XMLTools.hh"

#include "MessageDialog.hh"

#include "main_window.hh"

/****************************************************************/

FWObjectDatabase*  FWObjectDatabase::db         = NULL;
Group*             FWObjectDatabase::ScratchPad = NULL;

const string FWObjectDatabase::DB_TEMPLATE_FILE_NAME = "objects_init.xml" ;
const string FWObjectDatabase::DTD_FILE_NAME         = "fwbuilder.dtd"    ;

const char* FWObjectDatabase::TYPENAME={"FWObjectDatabase"};

long FWObjectDatabase::IDcounter = 0;

FWObjectDatabase::FWObjectDatabase() : FWObject()
{
    assert(db==NULL);
    db=this;

    IDcounter=time(NULL);

    data_file="";
    setName(TYPENAME);

    ScratchPad        = new Group();
    ScratchPad->setName("ScratchPad");

    setDirty(false);
}

FWObjectDatabase::FWObjectDatabase(const string &filename) throw(FWException)
:FWObject()
{
    assert(db==NULL);
    db=this;

    IDcounter=time(NULL);

    data_file="";
    setName(TYPENAME);

    ScratchPad        = new Group();
    ScratchPad->setName("ScratchPad");

    setDirty(false);
    load(filename);
}

FWObjectDatabase::~FWObjectDatabase()
{
}

void FWObjectDatabase::setFileName(const string &filename)
{
    data_file = filename;
}

const string &FWObjectDatabase::getFileName() 
{ 
    return data_file; 
}

const string FWObjectDatabase::getFileDir() 
{ 
    string::size_type i = data_file.rfind('/');
    if (i==string::npos || i==0) 
	return "";
    else 
	return data_file.substr(0,i);
}

bool FWObjectDatabase::saveIfModified() throw(FWException)
{
    MessageDialog::DlgReturnValue action;

    action=MessageDialog::Question(
"Some objects have been modified but not saved.
Do you want to save changes now ?");

    switch (action) 
    {
    case MessageDialog::OK_YES: save(); break; 
    case MessageDialog::NO:             break;         
    case MessageDialog::CANCEL:         return(false);
    }
    return(true);
}

void FWObjectDatabase::load(const string &f) throw(FWException)
{
    if(isDirty(true)) 
    {   // checking all the object recursively
	if (!saveIfModified()) 
            return;  // "Cancel"
    }
    
    if(f=="") 
    {
	FileSel *fs=new FileSel("Select file",
				Preferences::global_prefs->getWdir(),
				"*.xml");
	string df=fs->run();
	delete fs;
	if (df=="") return;
	data_file = df;
    } else
	data_file = f;
    
    if(data_file=="") 
        return;
    
    xmlDocPtr doc=XMLTools::loadFile(data_file, FWObjectDatabase::TYPENAME, FWObjectDatabase::DTD_FILE_NAME);
    
    xmlNodePtr root = xmlDocGetRootElement(doc);
    
    if(!root || !root->name || strcmp(FROMXMLCAST(root->name), FWObjectDatabase::TYPENAME)!=SAME)
    {
	xmlFreeDoc(doc);
        throw FWException("Data file have invalid structure: "+data_file);
    }

    /*
     * clear scratch pad: delete all children of ScratchPad 
     */
    ScratchPad->clearChildren();

    clearChildren();

    fromXML(root);

    std.AnyNetworkId    = "sysid0";
    std.AnyIPServiceId  = "sysid1";
    std.AnyIntervalId   = "sysid2";

    std.ObjectsId       = "stdid01";
    std.HostsId         = "stdid02";
    std.NetworksId      = "stdid03";
    std.ObjectGroupsId  = "stdid04";
    std.ServicesId      = "stdid05";
    std.IPServicesId    = "stdid06";
    std.ICMPServicesId  = "stdid07";
    std.UDPServicesId   = "stdid08";
    std.TCPServicesId   = "stdid09";
    std.ServiceGroupsId = "stdid10";
    std.TimeId          = "stdid11";
    std.FirewallsId     = "stdid12";

    FWObjectDatabase::db->copyToScratchPad();

    setDirty(false,true);  // clear dirty flag for all objects, recursive oper.

    if (main_w)
	main_w->showStatusbarMsg(string("Loaded data: ")+data_file);
}

void FWObjectDatabase::save() throw(FWException)
{
    if (data_file=="") 
    {
	saveAs();
	return;
    }

    saveFile(data_file);
}

void FWObjectDatabase::saveAs() throw(FWException)
{
    string df;

    FileSel *fs=new FileSel("Select file",
			    Preferences::global_prefs->getWdir(),
			    "*.xml");
    df=fs->run();
    delete fs;
    if (df=="") return;
    data_file=df;

    saveFile(data_file);
}

void FWObjectDatabase::saveXML(xmlDocPtr doc) throw(FWException)
{
    // create root node
    doc->children = xmlNewDocNode(doc, NULL, STRTOXMLCAST(getName()), NULL);
    toXML(xmlDocGetRootElement(doc));
    XMLTools::setDTD(doc, FWObjectDatabase::TYPENAME, FWObjectDatabase::DTD_FILE_NAME);
}

void FWObjectDatabase::saveFile(const string &filename) throw(FWException)
{
/*
    if (main_w)
	main_w->ClearRightPane(true);
*/

    xmlDocPtr doc = xmlNewDoc(TOXMLCAST("1.0"));
  
    doc->children = xmlNewDocNode(doc, NULL, STRTOXMLCAST(getName()), NULL);
    toXML(xmlDocGetRootElement(doc));
    
    XMLTools::saveFile(doc, filename, FWObjectDatabase::TYPENAME, FWObjectDatabase::DTD_FILE_NAME);
    
    xmlFreeDoc(doc);  
    setDirty(false,true);  // clear dirty flag for all objects, recursive oper.
    
    if(main_w)
	main_w->showStatusbarMsg(string("Saved data as: ")+filename);
}

xmlNodePtr FWObjectDatabase::toXML(xmlNodePtr parent)
{
    FWObject *o;

    xmlNewProp(parent,NULL,NULL);

    xmlNewProp(parent, 
               TOXMLCAST("version") , 
               TOXMLCAST(VERSION));

    vector<FWObject*>::iterator j;
    for(j=begin(); j!=end(); ++j) 
    {
        if ((o=(*j))!=NULL)   o->toXML(parent);
    }
    
    return parent;
}

/*
 *  This method finds a scratch pad copy of object "obj"
 */
FWObject* FWObjectDatabase::getScratchCopy(FWObject *obj)
{
    if (obj==FWObjectDatabase::db) return NULL;

    FWObject *p=FWObjectDatabase::ScratchPad->getById( obj->getId(), true );
    return p;
}

void FWObjectDatabase::copyToScratchPad()
{
    FWObject *p;
    FWObject *o;

    ScratchPad->clearChildren();

    vector<FWObject*>::iterator j;
    for(j=begin(); j!=end(); ++j)     {
        if((p=(*j))!=NULL)   {

	    o = FWObjectDatabase::create( p->getTypeName() );
	    *o = *p;
	    FWObjectDatabase::ScratchPad->add( o );
	}
    }
}

FWObject* FWObjectDatabase::create(const string& type_name)
{
    const char *n=type_name.c_str();
    FWObject   *nobj;

    if(strcmp(Policy::TYPENAME,n)==SAME)    
    { nobj = new Policy();
    } else if(strcmp(InterfacePolicy::TYPENAME,n)==SAME)    
    { nobj = new InterfacePolicy();
    } else if(strcmp(NAT::TYPENAME,n)==SAME)    
    { nobj = new NAT();
    } else if(strcmp(PolicyRule::TYPENAME,n)==SAME)    
    { nobj = new PolicyRule();
    } else if(strcmp(NATRule::TYPENAME,n)==SAME)    
    { nobj = new NATRule();
    } else if (strcmp(RuleElementSrc::TYPENAME,n)==SAME)
    { nobj = new RuleElementSrc();
    } else if (strcmp(RuleElementDst::TYPENAME,n)==SAME)      
    { nobj = new RuleElementDst();
    } else if (strcmp(RuleElementSrv::TYPENAME,n)==SAME)
    { nobj = new RuleElementSrv();
    } else if (strcmp(RuleElementOSrc::TYPENAME,n)==SAME)
    { nobj = new RuleElementOSrc();
    } else if (strcmp(RuleElementODst::TYPENAME,n)==SAME)
    { nobj = new RuleElementODst();
    } else if (strcmp(RuleElementOSrv::TYPENAME,n)==SAME)
    { nobj = new RuleElementOSrv();
    } else if (strcmp(RuleElementTSrc::TYPENAME,n)==SAME)
    { nobj = new RuleElementTSrc();
    } else if (strcmp(RuleElementTDst::TYPENAME,n)==SAME)
    { nobj = new RuleElementTDst();
    } else if (strcmp(RuleElementTSrv::TYPENAME,n)==SAME)
    { nobj = new RuleElementTSrv();
    } else if (strcmp(RuleElementInterval::TYPENAME,n)==SAME)
    { nobj = new RuleElementInterval();
    } else if(strcmp(FirewallOptions::TYPENAME,n)==SAME)
    { nobj=new FirewallOptions();
    } else if(strcmp(PolicyRuleOptions::TYPENAME,n)==SAME)
    { nobj=new PolicyRuleOptions();
    } else if(strcmp(NATRuleOptions::TYPENAME,n)==SAME)
    { nobj=new NATRuleOptions();
    } else if(strcmp(Host::TYPENAME,n)==SAME)
    { nobj=new Host();
    } else if(strcmp("AnyNetwork",n)==SAME)
    {
	nobj = new Network();
        nobj->setXMLName("AnyNetwork");
    } else if(strcmp("AnyIPService",n)==SAME)    
    { nobj=new IPService();
        nobj->setXMLName("AnyIPService");
    } else if(strcmp("AnyInterval",n)==SAME)    
    { nobj=new Interval();
        nobj->setXMLName("AnyInterval");
    } else if(strcmp(Firewall::TYPENAME,n)==SAME)    
    { nobj=new Firewall();
    } else if(strcmp(Network::TYPENAME,n)==SAME)    
    { nobj=new Network();
    } else if(strcmp(Interface::TYPENAME,n)==SAME)    
    { nobj=new Interface();
    } else if(strcmp(FWObjectReference::TYPENAME,n)==SAME)    
    { nobj=new FWObjectReference();
    } else if(strcmp(FWServiceReference::TYPENAME,n)==SAME)    
    { nobj=new FWServiceReference();
    } else if(strcmp(FWIntervalReference::TYPENAME,n)==SAME)    
    { nobj=new FWIntervalReference();
    } else if(strcmp(TCPService::TYPENAME,n)==SAME)    
    { nobj=new TCPService();
    } else if(strcmp(UDPService::TYPENAME,n)==SAME)    
    { nobj=new UDPService();
    } else if(strcmp(ICMPService::TYPENAME,n)==SAME)    
    { nobj=new ICMPService();
    } else if(strcmp(IPService::TYPENAME,n)==SAME)    
    { nobj=new IPService();
    } else if(strcmp(Interval::TYPENAME,n)==SAME)    
    { nobj=new Interval();
    } else if(strcmp(ObjectGroup::TYPENAME,n)==SAME)
    { nobj = new ObjectGroup();   
    } else if(strcmp(ServiceGroup::TYPENAME,n)==SAME)
    { nobj = new ServiceGroup();  
    } else if(strcmp(IntervalGroup::TYPENAME,n)==SAME)
    { nobj = new IntervalGroup();
    } else 
    {   
        //TODO: throw exception
        cerr << "unknown node: " << n << endl;
        return NULL;
    }

//    nobj->setName(n);

    return(nobj);
}


/**
 * Creates instance of FWObject depending on node type 
 */
FWObject *FWObjectDatabase::createFromXML(xmlNodePtr data)
{
    FWObject *nobj;
    const char *n=FROMXMLCAST(data->name);
    if (!n) return NULL;

    nobj=create(n);
    
    return(nobj);
}

string FWObjectDatabase::generateUniqueId()
{
    char res[20];

    sprintf(res,"id%lX",IDcounter);
    IDcounter++;

    return res;
}

/*
 *  Any object can be a child for the database
 */
bool FWObjectDatabase::validateChild(FWObject *o)
{ 
    return true;
}

