/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: IPAddress.hh,v 1.19 2001/07/23 21:54:22 vkurland Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _IPAddress_HH
#  define _IPAddress_HH

#include <string>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef ulong
/* header files on FreeBSD do not seem to define ulong anymore, or we include
 * them in a wrong order 
 */
typedef unsigned long int ulong;
#endif

#include "FWException.hh"

class IPAddress
{
    protected:

    friend class IPNetwork;

    unsigned int octets[4];
    virtual void validate() throw(FWException);
    
    public:
    
    explicit IPAddress();

    IPAddress(const unsigned char *data, size_t len) throw(FWException);
    IPAddress(const struct in_addr *) throw(FWException);
    explicit IPAddress(const string &) throw(FWException, FWNotSupportedException);
    IPAddress(const IPAddress &);

    IPAddress& operator=(const IPAddress &);

    IPAddress& operator=(const string &s) throw(FWException, FWNotSupportedException);

    friend bool operator<(const IPAddress &a, const IPAddress &b);
    friend bool operator==(const IPAddress &a, const IPAddress &b);
    unsigned const int operator[](size_t pos) const { return octets[pos]; }
    
    operator string() const { return toString(); }
    operator ulong() const;
    
    string toString() const;
   
};

class Netmask: public IPAddress
{
    protected:
    
    virtual void validate() throw(FWException);

    public:

    explicit Netmask();
    explicit Netmask(const string &) throw(FWException);
    Netmask(const unsigned char *data, size_t len) throw(FWException);

    /**
     * Creates natural classful netmask for given IP address.
     */
    explicit Netmask(const IPAddress &);

    Netmask& operator=(const string &s) throw(FWException);

    friend bool operator==(const Netmask &a, const Netmask &b);

};

class IPNetwork
{
    private:
    
    IPAddress address;
    Netmask   netmask;

    public:

    IPNetwork(const IPAddress &, const Netmask&);

    const IPAddress& getAddress() const;
    const Netmask&   getNetmask() const;

    IPNetwork& operator=(const IPNetwork &o);
    friend bool operator==(const IPNetwork &a, const IPNetwork &b);
    friend bool operator<(const IPNetwork &a, const IPNetwork &b);
    operator string() const { return address.toString()+"/"+netmask.toString(); }

    bool belongs(const IPAddress &) const;
    
};

class IPRoute
{
    public:

    IPRoute(const IPAddress &_dst, const Netmask &_nm, bool _direct)
    {
        nm     = _nm     ;
        dst    = _dst    ;
        direct = _direct ;
    }

    bool isDirect() const { return direct; }
    const Netmask   &getNetmask    () const { return nm;  }
    const IPAddress &getDestination() const { return dst; }
    
    private:
    
    Netmask   nm     ;
    IPAddress dst    ;
    bool      direct ;
};

#endif

