/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: Rule.cc,v 1.25 2001/07/19 07:43:25 vkurland Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "FWObjectDatabase.hh"

#include "Rule.hh"
#include "helpers.hh"
#include "Policy.hh"
#include "FWOptions.hh"

#include "XMLTools.hh"

#include "Firewall.hh"
#include "InterfacePolicy.hh"
#include "Policy.hh"
#include "MessageDialog.hh"
 
const char *Rule::TYPENAME={"Rule"};

Rule::Rule()
{
    setInt("position",0);

    enable();
}


void Rule::setPosition(int n)
{
    setInt("position",n);
}

int  Rule::getPosition()
{
    return getInt("position");
}

FWOptions* Rule::getOptionsObject()
{
    return NULL;
}

/**
 *   this method creates rule element and adds it to the rule as a
 *   child. It also pulls from resources and assigns to the rule
 *   element a references to its "any" object or default value
 * 
 *   TODO:  throw exception on errors (if could not find resources)
 *
 */
void Rule::createAndAddRuleElement(const char* element_name)
{
    FWObject    *re;

    re=FWObjectDatabase::create(element_name);
    assert (re!=NULL);

    add(re);

    if ( RuleElement::cast(re)!=NULL )
	(RuleElement::cast(re))->setAnyElement();

}



void Rule::disable()
{
    setBool("disabled",true);
}

void Rule::enable()
{
    setBool("disabled",false);
}

bool Rule::isDisabled()
{
    return( getBool("disabled") );
}

/*
 *************************************************************************
 */


const char *PolicyRule::TYPENAME={"PolicyRule"};

PolicyRule::PolicyRule()
{
    setStr("action","Deny");
    setBool("log",true);
}

void PolicyRule::fromXML(xmlNodePtr root)
{
    const char* n;

    FWObject::fromXML(root);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("position")));
    if(n)  setInt("position",atoi(n));

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("disabled")));
    if(n)  setStr("disabled",n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("action")));
    if(n)  setStr("action",n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("log")));
    if(n)  setStr("log",n);


    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("interface")));
    if(n)  setStr("interface",n);

    n= FROMXMLCAST(xmlGetProp(root,TOXMLCAST("direction")));
    if(n)  setStr("direction",n);

/*
 *  old versions did not create RuleOptions for Interface Policy Rules
 */
    FWObject *opt=getFirstByType( PolicyRuleOptions::TYPENAME );
    if (opt==NULL) {
	add( new PolicyRuleOptions() );
    }


    FWObject *fw, *p;

    fw=this;
    while (fw && !Firewall::isA(fw))
	fw=fw->getParent();

    assert(fw!=NULL);

    p=getParent();
    if (InterfacePolicy::isA(p)) {
// no restrictions for InterfacePolicy (as of yet)
    }

    if (Policy::isA(p)) {
	if ( exists("interface") ) {
	    string str;
	    str =string("Firewall: ") + fw->getName() + ",  Policy rule # " 
		+ getStr("position") + 
": Unsupported feature: Interface specified ('" + getStr("interface") + "')

Firewall Builder does not support interface specification 
in general firewall policy rules. Rules which require
interface specification must be placed in the policy associated
with corresponding firewall interface. 

Rule will be loaded now, but resulting firewall policy may be incorrect. 
Please verify it.";

	    MessageDialog::Warning(str);

	    remStr("interface");
	}

	if ( exists("direction") ) {
	    string str;
	    str =string("Firewall: ") + fw->getName() + ",  Policy rule # " 
		+ getStr("position") + 
": Unsupported feature: Direction specified ('" + getStr("direction") + "')

Firewall Builder does not support direction specification 
in general firewall policy rules. Rules which require
direction specification must be placed in the policy associated
with firewall interface. 

Rule will be loaded now, but resulting firewall policy may be incorrect. 
Please verify it.";


	    MessageDialog::Warning(str);

	    remStr("direction");
	}
    }

/*
 *   TODO:  remove this fragment when all data conversion is over
 */
    FWObject *srv=getFirstByType( RuleElementSrv::TYPENAME );
    if (srv==NULL) 
	createAndAddRuleElement(RuleElementSrv::TYPENAME );

}

xmlNodePtr PolicyRule::toXML(xmlNodePtr parent)
{
    xmlNodePtr me = xmlNewChild(parent, NULL, 
				STRTOXMLCAST(getTypeName()), 
				NULL);

    FWObject *o;
    
    map<string, string>::iterator i;
    for(i=data.begin(); i!=data.end(); ++i) 
    {
        string name  = (*i).first;
        string value = (*i).second;

        xmlAttrPtr pr = xmlNewProp(me, 
                                   STRTOXMLCAST(name) , 
                                   STRTOXMLCAST(value));
        if(name=="id")
        {
            xmlAddID(NULL, parent->doc, STRTOXMLCAST(value), pr);
        } else if(name=="ref") 
        {
            xmlAddRef(NULL, parent->doc, STRTOXMLCAST(value), pr);
        }
    }
    
    if ( (o=getFirstByType( RuleElementSrc::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementDst::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementSrv::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementInterval::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( PolicyRuleOptions::TYPENAME ))!=NULL )
	o->toXML(me);

    return me;
}

FWOptions* PolicyRule::getOptionsObject()
{
    return FWOptions::cast( getFirstByType(PolicyRuleOptions::TYPENAME) );
}


/*
 *************************************************************************
 */


const char *NATRule::TYPENAME={"NATRule"};

void NATRule::fromXML(xmlNodePtr root)
{
    const char* n;

    FWObject::fromXML(root);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("disabled")));
    if(n)  setStr("disabled",n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("position")));
    if(n)  setStr("position",n);
}

xmlNodePtr NATRule::toXML(xmlNodePtr parent)
{
    xmlNodePtr me = xmlNewChild(parent, NULL, 
				STRTOXMLCAST(getTypeName()), 
				NULL);

    FWObject *o;
    
    map<string, string>::iterator i;
    for(i=data.begin(); i!=data.end(); ++i) 
    {
        string name  = (*i).first;
        string value = (*i).second;

        xmlAttrPtr pr = xmlNewProp(me, 
                                   STRTOXMLCAST(name) , 
                                   STRTOXMLCAST(value));
        if(name=="id")
        {
            xmlAddID(NULL, parent->doc, STRTOXMLCAST(value), pr);
        } else if(name=="ref") 
        {
            xmlAddRef(NULL, parent->doc, STRTOXMLCAST(value), pr);
        }
    }
    
    if ( (o=getFirstByType( RuleElementOSrc::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementODst::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementOSrv::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementTSrc::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementTDst::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementTSrv::TYPENAME ))!=NULL )
	o->toXML(me);

    if ( (o=getFirstByType( RuleElementInterval::TYPENAME ))!=NULL )
	o->toXML(me);


    if ( (o=getFirstByType( NATRuleOptions::TYPENAME ))!=NULL )
	o->toXML(me);

    return me;
}

FWOptions* NATRule::getOptionsObject()
{
    return FWOptions::cast( getFirstByType(NATRuleOptions::TYPENAME) );
}














