/* 

                          Firewall Builder

                 Copyright (C) 2001 Vadim Zaliva, Vadim Kurland

  Author:  Vadim Zaliva lord@crocodile.org

  $Id: dns.hh,v 1.26 2001/07/17 19:06:09 vkurland Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifndef _DNS_HH_
#define _DNS_HH_

#include "config.h"

#include "IPAddress.hh"
#include "FWException.hh"
#include "BackgroundOp.hh"
#include "ThreadTools.hh"

#include <vector>
#include <map>
#include <queue>

#include <netdb.h>

#include <resolv.h>

class HostEnt
{
    public:
    
    string      name    ;
    set<string> aliases ;
};

/**
 * This is abstract class
 */
class DNS : public BackgroundOp
{
    public:

    DNS::DNS();

    /**
     * Finds IP adddresses of the host with given host name.
     * This operation does not run in backgound.
     */
    static vector<IPAddress> getHostByName(const string &name) throw(FWException);

    /**
     * Find all host names of host with given IP.
     * This operation does not run in backgound.
     */
    static HostEnt getHostByAddr(const IPAddress &addr) throw(FWException);

    /**
     * Return name of host with given IP.
     */
    static HostEnt getHostByAddr(const IPAddress &addr, int retries_, int timeout_) throw(FWException);

    /**
     * Finds NS records for given domain
     */
    multimap<string, IPAddress> getNS(const string &domain, Logger *logger, int retries_=RES_DFLRETRY, int timeout_=RES_TIMEOUT) throw(FWException);

    /**
     * Attempts to fetch zone for given domain for
     * it's name server. If succeeded, finds all 'A'
     * records in it and returs hostname/IPAddress pairs.
     */
    map<string, set<IPAddress> > findA(const string &domain, Logger *logger,  int retries_=RES_DFLRETRY, int timeout_=RES_TIMEOUT) throw(FWException);

    /**
     * Attempts to fetch zone for given domain for
     * specific server. If succeeded, finds all 'A'
     * records in it and returs hostname/IPAddress pairs.
     */
    map<string, set<IPAddress> > findA(const string &domain, const IPAddress &ns, Logger *logger,  int retries_=RES_DFLRETRY, int timeout_=RES_TIMEOUT) throw(FWException);

    protected:

    static string getErrorMessage(int rcode);

    private:

    static const size_t RSP_BUF_SIZE;

};

class DNS_findA_query : public DNS
{
    public:

    explicit DNS_findA_query();

    DNS_findA_query(const string &domain_, const IPAddress &ns_, int retries_=RES_DFLRETRY, int timeout_=RES_TIMEOUT);

    void init(const string &domain_, const IPAddress &ns_, int retries_=RES_DFLRETRY, int timeout_=RES_TIMEOUT);
    
    virtual void run_impl(Logger *logger) throw(FWException);
    
    map<string, set<IPAddress> > getResult() { return result; }

    private:

    map<string, set<IPAddress> > result;    
    string domain;
    int retries;
    int timeout;
    IPAddress ns;
};

class DNS_getNS_query : public DNS
{
    public:

    DNS_getNS_query(const string &domain_, int retries_=RES_DFLRETRY, int timeout_=RES_TIMEOUT);
    
    virtual void run_impl(Logger *logger) throw(FWException);
    
    multimap<string, IPAddress> getResult() { return result; }
    
    private:
    
    multimap<string, IPAddress> result;    
    string domain;
    int retries;
    int timeout;
};

class DNS_bulkBackResolve_query : public DNS
{
    friend void *DNS_bulkBackResolve_Thread(void *);

    public:
    
    DNS_bulkBackResolve_query(set<IPAddress>, unsigned int nthreads,  
			      int retries_=RES_DFLRETRY,
			      int timeout_=RES_TIMEOUT);
    virtual ~DNS_bulkBackResolve_query();
    
    /**
     * Resolves all hosts in list. If resolving at least of one
     * of them failed - exception is thrown.
     * Even when exception is thrown, successfully resolved hosts
     * are returned.
     */
    virtual void run_impl(Logger *logger) throw(FWException);
    
    map<IPAddress, HostEnt> getResult() { return result; }
    set<IPAddress> getFailed() { return failed; }
    
    protected:
    
    map<IPAddress, HostEnt> result;
    Mutex failed_mutex;
    
    set<IPAddress> failed;
    Mutex result_mutex;
    
    queue<IPAddress> ips;
    Mutex queue_mutex;

    int retries;
    int timeout;
    
    private:

    GMutex *running_mutex       ;
    GCond  *running_cond        ;
    unsigned int runnning_count ;
    
    unsigned int nthreads;
};

#endif



