/* 

                          Firewall Builder

                 Copyright (C) 2001 Vadim Zaliva, Vadim Kurland

  Author:  Vadim Zaliva lord@crocodile.org

  $Id: snmp.cc,v 1.48 2001/07/19 13:58:19 lord Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

#ifdef HAVE_LIBSNMP

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include "helpers.hh"
#include "snmp.hh"

#include <typeinfo>
#include <algorithm>

const char *SNMPQuery::SNMP_INTERFACES_NUM   = "interfaces.ifNumber.0"                     ;
const char *SNMPQuery::SNMP_INTERFACE_INDEX  = "interfaces.ifTable.ifEntry.ifIndex"        ;
const char *SNMPQuery::SNMP_INTERFACES_DESCR = "interfaces.ifTable.ifEntry.ifDescr"        ;
const char *SNMPQuery::SNMP_INTERFACES_PHYSA = "interfaces.ifTable.ifEntry.ifPhysAddress"  ;
const char *SNMPQuery::SNMP_ADDR_INDEX_TABLE = "ip.ipAddrTable.ipAddrEntry.ipAdEntIfIndex" ;
const char *SNMPQuery::SNMP_NMASK_TABLE      = "ip.ipAddrTable.ipAddrEntry.ipAdEntNetMask" ;
const char *SNMPQuery::SNMP_ADDR_TABLE       = "ip.ipAddrTable.ipAddrEntry.ipAdEntAddr"    ;
const char *SNMPQuery::SNMP_AT_TABLE_NET     = "at.atTable.atEntry.atNetAddress";
const char *SNMPQuery::SNMP_AT_TABLE_PHYS    = "at.atTable.atEntry.atPhysAddress";

const char *SNMPQuery::SNMP_SYSNAME          = "system.sysName.0"     ;
const char *SNMPQuery::SNMP_SYSDESCR         = "system.sysDescr.0"    ;
const char *SNMPQuery::SNMP_CONTACT          = "system.sysContact.0"  ;
const char *SNMPQuery::SNMP_LOCATION         = "system.sysLocation.0" ;

const char *SNMPQuery::SNMP_ROUTE_DST_TABLE  = "ip.ipRouteTable.ipRouteEntry.ipRouteDest" ;
const char *SNMPQuery::SNMP_ROUTE_NM_TABLE   = "ip.ipRouteTable.ipRouteEntry.ipRouteMask" ;
const char *SNMPQuery::SNMP_ROUTE_TYPE_TABLE = "ip.ipRouteTable.ipRouteEntry.ipRouteType" ;
const int   SNMPQuery::SNMP_DIRECT_ROUTE     = 3;

#include <iostream>


SNMPQuery::SNMPQuery(string h,string c, int retries_, long timeout_)
{
    hostname  = h  ;
    community = c  ;
    retries   = retries_;
    timeout   = timeout_;

    descr     = "" ;
    contact   = "" ;
    location  = "" ;
}

void SNMPQuery::init(string h, string c, int retries_, long timeout_)
{
    hostname  = h  ;
    community = c  ;
    retries   = retries_;
    timeout   = timeout_;

    descr     = "" ;
    contact   = "" ;
    location  = "" ;
}

SNMPQuery::~SNMPQuery()
{
}

void SNMPQuery::fetchAll(Logger *logger) throw(FWException)
{
    if(community.empty())
        throw FWException("No SNMP community specified");

    if(hostname.empty())
        throw FWException("No SNMP hostname specified");

    SNMPConnection c(hostname, community);
    *logger << "Connecting to " << hostname << "\n";
    c.connect(retries, timeout);

    check_stop();
    
    fetchSysInfo   (logger, &c);
    check_stop();
    
    fetchInterfaces(logger, &c);
    check_stop();
    
    fetchArpTable  (logger, &c);
    check_stop();

    fetchRoutingTable  (logger, &c);
    check_stop();
}

void SNMPQuery::fetchArpTable(Logger *logger, SNMPConnection *connection) throw(FWException)
{
    *logger << "Fetching ARP table\n"; 
    arptable.clear();
    
    SNMPConnection *c;
    if(connection)
    {
        c=connection;
    } else
    {
        if(community.empty())
            throw FWException("No SNMP community specified");
       
        if(hostname.empty())
            throw FWException("No SNMP hostname specified");
       
        check_stop();
        c=new SNMPConnection(hostname, community);
        *logger << "Connecting to " << hostname << "\n";
        try
        {
            c->connect(retries, timeout);
        } catch(...)
        {
            // If program was stopped, show this, rather
            // than error.
            check_stop();
            throw;
        }
        check_stop();
    }

    vector<SNMPVariable*> v;
    try
    {
        multimap<string, SNMPVariable* > nw=c->walk(SNMP_AT_TABLE_NET  );
        multimap<string, SNMPVariable* > pw=c->walk(SNMP_AT_TABLE_PHYS );
        try
        {
            check_stop();
            int preflen = strlen(SNMP_AT_TABLE_NET);
            for(multimap<string, SNMPVariable* >::iterator j = nw.begin(); j!=nw.end(); ++j)
            {
                check_stop();
                if((*j).second->type!=SNMPVariable::snmp_ipaddr)
                {
                    *logger << "unexpected result type in '" << 
                    SNMP_AT_TABLE_NET << "'table. Skipping it.\n";
                    continue;
                }
                IPAddress ip = dynamic_cast<SNMPVariable_IPaddr*>((*j).second)->getIPAddressValue();
                string suffix = ((*j).first).substr(preflen);
                string hwaddr="?";
                
                vector<SNMPVariable*> v = c->get(string(SNMP_AT_TABLE_PHYS)+suffix);
                if(v.size()==1)
                    if(v[0]->type==SNMPVariable::snmp_string)
                        hwaddr=dynamic_cast<SNMPVariable_String*>(v[0])->toHexString();
                
                SNMPVariable::freeVarList(v);
                arptable.insert(make_pair(ip,hwaddr));
            }
        } catch(...)
        {
            for(multimap<string, SNMPVariable* >::iterator j = nw.begin(); j!=nw.end(); ++j)
                delete (*j).second;
            for(multimap<string, SNMPVariable* >::iterator j = pw.begin(); j!=pw.end(); ++j)
                delete (*j).second;
            throw;
        }
        for(multimap<string, SNMPVariable* >::iterator j = nw.begin(); j!=nw.end(); ++j)
            delete (*j).second;
        for(multimap<string, SNMPVariable* >::iterator j = pw.begin(); j!=pw.end(); ++j)
            delete (*j).second;
    } catch (FWException &ex)
    {
        SNMPVariable::freeVarList(v);
        if(!connection)
            delete c;
        // If program was stopped, show this, rather
        // than error.
        check_stop();
        throw;
    }
    if(!connection)
        delete c;
    *logger << "Done fetching ARP table\n";
}

void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) throw(FWException)
{
    *logger << "Fetching routing table\n"; 
    routes.clear();
    
    SNMPConnection *c;
    if(connection)
    {
        c=connection;
    } else
    {
        if(community.empty())
            throw FWException("No SNMP community specified");
        
        if(hostname.empty())
            throw FWException("No SNMP hostname specified");
        
        check_stop();
        c=new SNMPConnection(hostname, community);
        *logger << "Connecting to " << hostname << "\n";
        try
        {
            c->connect(retries, timeout);
        } catch(...)
        {
            // If program was stopped, show this, rather
            // than error.
            check_stop();
            throw;
        }
        check_stop();
    }
    
    vector<SNMPVariable*> v;
    try
    {
        // Get Indices and destinations
        
        check_stop();
        multimap<string, SNMPVariable* > w=c->walk(SNMP_ROUTE_DST_TABLE);
        try
        {
            check_stop();
            for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
            {
                check_stop();
                if((*j).second->type!=SNMPVariable::snmp_ipaddr)
                {
                    *logger << "unexpected result type in '" << 
                    SNMP_ROUTE_DST_TABLE << "'table. Skipping it.\n";
                    continue;
                }
                IPAddress dst = dynamic_cast<SNMPVariable_IPaddr*>((*j).second)->getIPAddressValue();
                string rname=(*j).first.substr(strlen(SNMP_ROUTE_DST_TABLE)+1);

                check_stop();
                v=c->get(string(SNMP_ROUTE_NM_TABLE)+"."+rname);
                
                
                if(v[0]->type!=SNMPVariable::snmp_ipaddr)
                {
                    *logger << "unexpected result type in '" << 
                    SNMP_ROUTE_NM_TABLE << "'table. Skipping it.\n";
                    SNMPVariable::freeVarList(v);
                    continue;
                }
                Netmask nm = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getNetmaskValue();
                SNMPVariable::freeVarList(v);

                check_stop();
                v=c->get(string(SNMP_ROUTE_TYPE_TABLE)+"."+rname);
                if(v[0]->type!=SNMPVariable::snmp_int)
                {
                    *logger << "unexpected result type in '" << 
                    SNMP_ROUTE_TYPE_TABLE << "'table. Skipping it.\n";
                    SNMPVariable::freeVarList(v);
                    continue;
                }
                long type = dynamic_cast<SNMPVariable_Int*>(v[0])->getIntValue();
                
                routes.push_back(IPRoute(dst,nm,type==SNMP_DIRECT_ROUTE));
            }
        } catch(...)
        {
            for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
                delete (*j).second;
            throw;
        }
        for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
            delete (*j).second;
        w.clear();

    } catch (FWException &ex)
    {
        SNMPVariable::freeVarList(v);
        if(!connection)
            delete c;
        // If program was stopped, show this, rather
        // than error.
        check_stop();
        throw;
    }
    if(!connection)
        delete c;
    *logger << "Done fetching routing table\n";
}



void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) throw(FWException)
{
    interfaces.clear();
    
    SNMPConnection *c;
    if(connection)
    {
        c=connection;
    } else
    {
        if(community.empty())
            throw FWException("No SNMP community specified");
        
        if(hostname.empty())
            throw FWException("No SNMP hostname specified");
        
        check_stop();
        c=new SNMPConnection(hostname, community);
        *logger << "Connecting to " << hostname << "\n";
        try
        {
            c->connect(retries, timeout);
        } catch(...)
        {
            // If program was stopped, show this, rather
            // than error.
            check_stop();
            throw;
        }
        check_stop();
    }
    
    int     ni  = 0;
    vector<SNMPVariable*> v;
    try
    {
        *logger << "Getting Number Of interfaces\n";
        v=c->get(SNMP_INTERFACES_NUM);
        ni=SNMPVariable::varList2Int(v);
        SNMPVariable::freeVarList(v);
        check_stop();

	*logger << "Number of interfaces reported by SNMP agent: " << ni << "\n";
        
	if(ni!=0) 
        {
            // Get indices
            vector<long> indices(ni);
            *logger << "Getting interface indices: " << ni << "\n";
            for(int i=0; i<ni; i++)
            {
                char str[1024];
                
                sprintf(str,"%s.%d", SNMP_INTERFACE_INDEX , i+1);
                v=c->get(str);
                indices[i]=SNMPVariable::varList2Int(v);
                SNMPVariable::freeVarList(v);
                check_stop();
            }

            // Get desriptions
            vector<string> descr(ni);
            *logger << "Getting interface descriptions: " << ni << "\n";
            for(int i=0; i<ni; i++)
            {
                char str[1024];
                
                //TODO: Q: does it mapped by index or by number?
                sprintf(str,"%s.%ld", SNMP_INTERFACES_DESCR , indices[i]);
                v=c->get(str);
                descr[i]=SNMPVariable::varList2String(v);
                SNMPVariable::freeVarList(v);
                check_stop();
            }

            // Get physical addresses
            vector<string> physa(ni);
            *logger << "Getting interface physical addresses: " << ni << "\n";
            for(int i=0; i<ni; i++)
            {
                char str[1024];
                
                //TODO: Q: does it mapped by index or by number?
                sprintf(str,"%s.%ld", SNMP_INTERFACES_PHYSA , indices[i]);
                v=c->get(str);
                if(v.size()!=1)
                    throw FWException(string("Unexpected response length for variable: ")+str);
                if(v[0]->type!=SNMPVariable::snmp_string)
                    throw FWException(string("Unexpected response type for: ")+str);
                
                physa[i]=dynamic_cast<SNMPVariable_String*>(v[0])->toHexString();
                SNMPVariable::freeVarList(v);
                check_stop();
            }
                
            // Get IPs
            vector<string> addr(ni);
            *logger << "Getting interface addresses indices.\n";
            check_stop();

            multimap<string, SNMPVariable* > w=c->walk(SNMP_ADDR_INDEX_TABLE);
            try
            {
                check_stop();
                for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
                {
                    check_stop();
                    if((*j).second->type!=SNMPVariable::snmp_int)
                    {
                        *logger << "unexpected result type in '" << 
                        SNMP_ADDR_INDEX_TABLE << "'table. Skipping it.\n";
                        continue;
                    }
                    long ind = dynamic_cast<SNMPVariable_Int*>((*j).second)->getIntValue();
                    vector<long>::iterator it=std::find(indices.begin(), indices.end(), ind);
                    if(it==indices.end())
                    {
                        *logger << "Reference to unexisting index " << ind << " in '" <<
                        SNMP_ADDR_INDEX_TABLE << "'table. Skipping it.\n";
                        continue;
                    }
                    int off=distance(indices.begin(), it);
                    string iname=(*j).first.substr(strlen(SNMP_ADDR_INDEX_TABLE)+1);
                    addr[off]=iname;
                }
                
            } catch(...)
            {
                for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
                    delete (*j).second;
                throw;
            }
            for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
                delete (*j).second;
            w.clear();

            // Get netmasks and build resulting list
            *logger << "Getting interface netmasks and addresses.\n";
            for(int i=0; i<ni; i++)
            {
                if(!addr[i].empty())
                {
                    check_stop();
                    v=c->get(string(SNMP_NMASK_TABLE)+"."+addr[i]);
                    if(v.size()!=1)
                        throw FWException("Can't get netmask");
                    if(v[0]->type!=SNMPVariable::snmp_ipaddr)
                        throw FWException("Wrong return type for netmask");
                    string nm = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getNetmaskValue().toString();
                    SNMPVariable::freeVarList(v);
                    check_stop();
                    
                    v=c->get(string(SNMP_ADDR_TABLE)+"."+addr[i]);
                    if(v.size()!=1)
                        throw FWException("Can't get IP address");
                    if(v[0]->type!=SNMPVariable::snmp_ipaddr)
                        throw FWException("Wrong return type for IP address");
                    string ad = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getIPAddressValue().toString();
                    SNMPVariable::freeVarList(v);
                    check_stop();
                    
                    interfaces.push_back(
                        Interface(descr[i], ad, nm, false, false, physa[i])
                    );
                }
            }
	}
    } catch (FWException &ex)
    {
        SNMPVariable::freeVarList(v);
        if(!connection)
            delete c;
        // If program was stopped, show this, rather
        // than error.
        check_stop();
        throw;
    }
    if(!connection)
        delete c;
    *logger << "Done fetching interfaces\n";
}

void SNMPQuery::fetchSysInfo(Logger *logger, SNMPConnection *connection) throw(FWException)
{
    descr     = "" ;
    contact   = "" ;
    location  = "" ;
    sysname   = "" ;

    SNMPConnection *c;
    if(connection)
    {
        c=connection;
    } else
    {
        if(community.empty())
        {
            check_stop();
            throw FWException("No SNMP community specified");
        }
        
        if(hostname.empty())
        {
            check_stop();
            throw FWException("No SNMP hostname specified");
        }
        
        check_stop();
        c=new SNMPConnection(hostname, community);
        *logger << "Connecting to " << hostname << "\n";
        try
        {
            c->connect(retries, timeout);
        } catch(...)
        {
            // If program was stopped, show this, rather
            // than error.
            check_stop();
            throw;
        }
        check_stop();
    }
    
    try
    {
        vector<SNMPVariable*> v;
        
        *logger << "Getting System name\n";
        v=c->get(SNMP_SYSNAME);
        sysname    = SNMPVariable::varList2String(v);
        SNMPVariable::freeVarList(v);
        check_stop();

        *logger << "Getting Description\n";
        v=c->get(SNMP_SYSDESCR);
        descr    = SNMPVariable::varList2String(v);
        SNMPVariable::freeVarList(v);
        check_stop();
        
        *logger << "Getting Location\n";
        v=c->get(SNMP_LOCATION);
        location = SNMPVariable::varList2String(v);
        SNMPVariable::freeVarList(v);
        check_stop();
        
        *logger << "Getting Contact Info\n";
        v=c->get(SNMP_CONTACT);
        contact  = SNMPVariable::varList2String(v);
        SNMPVariable::freeVarList(v);
        check_stop();
    } catch(...)
    {
        if(!connection)
            delete c;
        // If program was stopped, show this, rather
        // than error.
        check_stop();
        throw;
    }
    if(!connection)
        delete c;
    *logger << "Done fetching sysinfo\n";
}

vector<IPRoute> SNMPQuery::getRoutes()
{
    return routes;
}

map<IPAddress, string> SNMPQuery::getArtpTable()
{
    return arptable;
}

vector<Interface> SNMPQuery::getInterfaces()
{
    return interfaces;
}

const string& SNMPQuery::getSysname()
{
    return sysname;
}

const string& SNMPQuery::getDescr()
{
    return descr;
}

const string& SNMPQuery::getContact()
{
    return contact;
}

const string& SNMPQuery::getLocation()
{
    return location;
}

// -----------------------------------------------

bool SNMPConnection::lib_initialized = false;

SNMPConnection::SNMPConnection(const string &p, const string &c)
{
    connected    = false ;
    session_data = NULL  ;
    peer         = p     ;
    community    = c     ;
    if(!lib_initialized)
    {
        init_snmp("fwbuilder");
        lib_initialized = true;
    }
}

SNMPConnection::~SNMPConnection()
{
    if(connected)
        disconnect();
}

void SNMPConnection::connect(int retries, long timeout) throw(FWException)
{
    if(connected)
        throw FWException("SNMPSession: already connected");
    
    session_data = new struct snmp_session;
    snmp_sess_init( session_data );
    session_data->version       = SNMP_VERSION_1;
    session_data->peername      = cxx_strdup(peer.c_str());
    session_data->community     = (unsigned char *)cxx_strdup(community.c_str());
    session_data->community_len = community.length();
    session_data->retries       = retries;
    session_data->timeout       = timeout;
    
    session = snmp_open(session_data);
    if(!session)
        throw FWException("SNMPSession: error while establishing connection.");
        
    connected=true;
}

void SNMPConnection::disconnect() throw(FWException)
{
    if(!connected)
        throw FWException("SNMPSession: already disconnected");

    snmp_close(session);

    delete session_data->peername;
    delete session_data->community;
    delete session_data;
    session_data = NULL;
    connected    = false;
}

multimap<string, SNMPVariable* > SNMPConnection::walk(const string &variable) throw(FWException)
{
    multimap<string, SNMPVariable*> res;

    oid    root[MAX_OID_LEN];
    size_t rootlen;
    
    rootlen = MAX_OID_LEN;
    get_node(variable.c_str(), root, &rootlen); //TODO: error check

    oid    name[MAX_OID_LEN];
    size_t name_length;

    /* get first object to start walk */
    memmove(name, root, rootlen * sizeof(oid));
    name_length = rootlen;

    bool running=true;
    while(running)
    {
        /* create PDU for GETNEXT request and add object name to request */
        struct snmp_pdu *pdu = snmp_pdu_create(SNMP_MSG_GETNEXT);
        snmp_add_null_var(pdu, name, name_length);
        
        /* do the request */
        struct snmp_pdu *response;
        int status = snmp_synch_response(session, pdu, &response);
        if(status == STAT_SUCCESS)
        {
            if(response->errstat == SNMP_ERR_NOERROR)
            {
                for(struct variable_list *vars = response->variables; 
                    vars; 
                    vars = vars->next_variable
                )
                {
                    if ((vars->name_length < rootlen) ||
                        (memcmp(root, vars->name, rootlen * sizeof(oid))!=0)) {
                        /* not part of this subtree */
                        running = false;
                        continue;
                    }

                    char n[MAX_NAME_LEN];
                    sprint_objid (n, vars->name, vars->name_length);
                    res.insert(make_pair(string(n),SNMPVariable::create(vars)));
                    
                    if ((vars->type != SNMP_ENDOFMIBVIEW) &&
                        (vars->type != SNMP_NOSUCHOBJECT) &&
                        (vars->type != SNMP_NOSUCHINSTANCE)){
                        /* not an exception value */
                        memmove((char *)name, (char *)vars->name,
                                vars->name_length * sizeof(oid));
                        name_length = vars->name_length;
                    } else
                    {
                        /* an exception value, so stop */
                        running = false;
                    }

                }
            } else 
            {
                /* error in response, print it */
                if(response->errstat == SNMP_ERR_NOSUCHNAME)
                {
                    throw FWException("SNMPSession: error: End of MIB");
                } else 
                {
                    throw FWException(string("SNMPSession: Error in packet. Reason:")+
                                      snmp_errstring(response->errstat));
                }
            }
        } else if (status == STAT_TIMEOUT)
        {
            throw FWException("SNMPSession: SNMP timeout");
        } else 
        {    /* status == STAT_ERROR */
            throw FWException("SNMPSession: SNMP error");
        }
        if (response)
            snmp_free_pdu(response);
    }
    
    return res;
}

vector<SNMPVariable*> SNMPConnection::get(const string &variable) throw(FWException)
{
    if(!connected)
        throw FWException("SNMPSession: not connected");
    
    struct snmp_pdu *pdu = snmp_pdu_create(SNMP_MSG_GET);
    oid anOID[MAX_OID_LEN];
    size_t anOID_len = MAX_OID_LEN;
    get_node(variable.c_str(), anOID, &anOID_len); //TODO: error check
    snmp_add_null_var(pdu, anOID, anOID_len);
    struct snmp_pdu *response;
    int status = snmp_synch_response(session, pdu, &response);
    if(status == STAT_SUCCESS && response->errstat == SNMP_ERR_NOERROR) 
    {
        vector<SNMPVariable*> res;
        for(struct variable_list *vars = response->variables; 
            vars; 
            vars = vars->next_variable
        )
        {
            try
            {
                res.push_back(SNMPVariable::create(vars));
            } catch(const FWException &ex)
            {
                SNMPVariable::freeVarList(res);
                throw;
            }
        }
        
        if (response)
            snmp_free_pdu(response);
        
        return res;
    } else
    {
        if (response)
            snmp_free_pdu(response);
        throw FWException("SNMPSession: Error getting variable "+ variable);
    } 
}

SNMPVariable *SNMPVariable::create(struct variable_list *vars) throw(FWException)
{
    switch(vars->type)
    {
    case ASN_INTEGER:
        return new SNMPVariable_Int(*vars->val.integer);
    case ASN_OCTET_STR:
        return new SNMPVariable_String(vars->val.string, vars->val_len);
    case ASN_BIT_STR:
        return new SNMPVariable_Bits(vars->val.bitstring, vars->val_len);
    case ASN_APP_COUNTER64:
        return new SNMPVariable_Counter64(vars->val.counter64);
    case ASN_OBJECT_ID:
        return new SNMPVariable_OID(*vars->val.objid);
    case ASN_IPADDRESS:
        return new SNMPVariable_IPaddr(vars->val.string, vars->val_len);
    default: 
        char x[32];
        sprintf(x,"%d", (int)vars->type);
        throw FWException(string("Unknown SNMP variable type: ")+x);
    }
}

string SNMPVariable_Int::toString()
{
    char x[32];
    sprintf(x,"%ld", value);
    return x;
}

string SNMPVariable_Bits::toString()
{
    string res="[";
    for(size_t i=0;i<len;i++)
    {
        char x[8];
        //TODO: now we print it in hex
        // we should print it in binary.
        sprintf(x,"%2x", (unsigned int)value[i]);
        res+=x;
    }
    res+="]";
    return res;
}

string SNMPVariable_IPaddr::toString()
{
    string res="SNMP IPAddress/Netmask[";
    for(size_t i=0;i<len;i++)
    {
        if(i) 
            res+=".";
        char x[8];
        sprintf(x,"%d", (unsigned int)value[i]);
        res+=x;
    }
    res+="]";
    return res;
}

IPAddress SNMPVariable_IPaddr::getIPAddressValue() throw(FWException)
{
    return IPAddress(value, len);
}

Netmask SNMPVariable_IPaddr::getNetmaskValue() throw(FWException)
{
    return Netmask(value, len);
}

string SNMPVariable_String::toString()
{
    return value;
}
const string SNMPVariable_String::toHexString()
{
    string res;
    for(unsigned i=0;i<value.length();i++)
    {
        if(i)
            res+=':';
        u_char c=value[i];
        char buf[16];
        sprintf(buf,"%02X",(unsigned int)c);
        res+=buf;
    }
    return res;
}


string SNMPVariable_Counter64::toString()
{
    char x[70];
    sprintf(x,"[%ld:%ld]", (long)low, (long)high);
    return x;
}

string SNMPVariable_OID::toString()
{
    char x[32];
    sprintf(x,"%ld", (long)value);
    return x;
}


void SNMPVariable::freeVarList(vector<SNMPVariable*> &v)
{
    vector<SNMPVariable*>::iterator j;
    for(j=v.begin(); j!=v.end(); ++j) 
        delete (*j);
    v.clear();
}

string SNMPVariable::varList2String(vector<SNMPVariable*> &v)
{
    string res;

    vector<SNMPVariable*>::iterator j;
    for(j=v.begin(); j!=v.end(); ++j) 
        res+=(*j)->toString();
    
    return res;
}

long SNMPVariable::varList2Int(vector<SNMPVariable*> &v) throw(FWException)
{
    if(v.size()!=1)
        throw FWException("Empty SNMP variable list returned. Could not extract integer");
    if(v[0]->type != SNMPVariable::snmp_int)
        throw FWException("Could not extract integer from non-int SNMP variable.");
    return dynamic_cast<SNMPVariable_Int *>(v[0])->getIntValue();
}

SNMPCrawler::SNMPCrawler()
{}

SNMPCrawler::SNMPCrawler(const IPAddress &_seed, 
                         const string &_community,
                         bool _recursive,
                         bool _skip_virtual,
                         bool _do_dns,
                         unsigned int _dns_threads,
                         int  _snmp_retries, 
                         long _snmp_timeout,
			 int  _dns_retries,
			 int  _dns_timeout,
                         const vector<IPNetwork> *_include)
{
    init(_seed, _community, _recursive, _skip_virtual, 
         _do_dns, _dns_threads, 
	 _snmp_retries,  _snmp_timeout, _dns_retries, _dns_timeout, 
	 _include);
}

SNMPCrawler::~SNMPCrawler()
{
}

void SNMPCrawler::init(const IPAddress &_seed, 
		       const string &_community,
		       bool _recursive,
                       bool _skip_virtual,
                       bool _do_dns,
                       unsigned int _dns_threads,
		       int _snmp_retries, 
		       long _snmp_timeout,
		       int  _dns_retries,
		       int  _dns_timeout,
		       const vector<IPNetwork> *_include)
{
    include      = _include      ;
    community    = _community    ;
    snmp_retries = _snmp_retries ;
    snmp_timeout = _snmp_timeout ;
    recursive    = _recursive    ;
    skip_virtual = _skip_virtual ;
    do_dns       = _do_dns       ;
    dns_threads  = _dns_threads  ;
    dns_retries  = _dns_retries  ;
    dns_timeout  = _dns_timeout  ;

    queue.clear();
    found.clear();
    networks.clear();

    queue[_seed]="";
}

bool SNMPCrawler::included(const IPAddress &a) const
{
    if(!include)
        return true; // no include list provided. All hosts are OK.
    
    for(vector<IPNetwork>::const_iterator i=include->begin(); i!=include->end(); ++i) {
        if((*i).belongs(a)) 
            return true;
    }
    return false;
}

bool SNMPCrawler::alreadyseen(const IPAddress &a) const
{
    return found.find(a) != found.end();
}

const IPNetwork SNMPCrawler::LOOPBACK_NET(IPAddress("127.0.0.0"),Netmask("255.0.0.0"));
const Netmask   SNMPCrawler::PTP_NETMASK("255.255.255.255");

bool SNMPCrawler::isvirtual(const IPAddress &addr, const string &pa) const 
{
    if(!pa.length())
        return false;

    for(map<IPAddress, CrawlerFind>::const_iterator i=found.begin(); i!=found.end(); ++i)
    {
        const CrawlerFind &c = (*i).second ;
        
        for(vector<Interface>::const_iterator j=c.interfaces.begin(); j!=c.interfaces.end(); ++j)
        {
            try
            {
                if(pa == (*j).getPhysicalAddress() && addr != (*j).getIPAddess())
                    return true;
            } catch(FWException &ex)
            {
                // Exception caused by interface having bad IP address.
                // we just skip this interface and go on
            }
        }
    }
    
    return false;
}

bool SNMPCrawler::point2point(const IPNetwork &n) const
{
    // TODO: it is not enough just to check netmask.
    return n.getNetmask()==PTP_NETMASK;
}

bool SNMPCrawler::special(const IPNetwork &n) const
{
    /*
     * Multicast :  224.0.0.0 - 239.0.0.0
     * Broadcast :  255.255.255.255  and  0.0.0.0,
     * loopback  :  All addresses on the net  127.0.0.0/255.0.0.0
     */
    
    IPAddress a=n.getAddress();
    
    if(a[0]>=224 && a[0]<=239 && a[1]==0 && a[2]==0 && a[3]==0)
        return true; // multicast
    
    if(a[0]==a[1] && a[1]==a[2] && a[2]==a[3] && (a[0]==255 || a[0]==0))
        return true; // broadcast
    
    if(n==LOOPBACK_NET)
        return true;

    return false;
}

//TODO: multiple threads (via pool).
void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
{
    *logger << "Starting SNMPCrawler. Seed host: " << (*(queue.begin())).first.toString() << "\n";
    SNMP_discover_query q;
    do {
        check_stop();
        map<IPAddress,string>::iterator i=queue.begin();
        if(i==queue.end())
        {
            break;
        }
        IPAddress task           = (*i).first  ;
        string task_phys_address = (*i).second ;
        queue.erase(i);
        
        check_stop();

        *logger << "Processing " << task.toString() << "\n";

        // Now in task we have element to probe
        
        q.init(string(task), // fake host make - IP in dotted notation
               community,
               snmp_retries,  
               snmp_timeout
        );
        
        check_stop();

        CrawlerFind *res=new CrawlerFind();
        found[task] = *res; delete res;
        found[task].found_phys_addr = task_phys_address ;
        try
        {
            q.fetchArpTable(logger);
            found[task].have_snmpd = true;
        } catch(const FWException &ex)
        {
            // fetch failed
            *logger << "Failed to fetch ARP table from " << task.toString() << "\n";
            continue;
        }
        check_stop();
        map<IPAddress, string> at=q.getArtpTable();    
        *logger << "Fetched " << long(at.size()) << " entries\n";

        int qplus=0, rplus=0, dplus=0;
        for(map<IPAddress, string>::iterator j=at.begin();j!=at.end();++j)
        {
            check_stop();

            IPAddress  c = (*j).first;
            string    pa = (*j).second;
            if(included(c) && !alreadyseen(c) && !isvirtual(c,pa))
            {
                if(recursive)
                {
                    qplus++;
                    queue[c]=pa;
                } else
                {
                    rplus++;
                    found[c]=CrawlerFind();
                    found[c].found_phys_addr=pa;
                }
            } else
                dplus++;
        }
        if(qplus) *logger << "Adding "   << qplus << " hosts to queue\n"   ;
        if(rplus) *logger << "Adding "   << rplus << " hosts to results\n" ;
        if(dplus) *logger << "Skipping " << dplus << " hosts as duplicate, excluded or virtual\n" ;
        
        try
        {
            q.fetchInterfaces(logger);
            found[task].have_snmpd = true;

            vector<Interface> intf = q.getInterfaces();
            found[task].interfaces=intf;
            for(vector<Interface>::iterator j=intf.begin(); j!=intf.end(); ++j)
            {
                IPNetwork net=(*j).getIPNetwork();
                if(!special(net)  && included(net.getAddress()) )
                {
                    *logger << "Network " << string(net) << " found.\n";
                    networks.insert(net);
                }
            }
            
        } catch(const FWException &ex)
        {
            // fetch failed
            *logger << "Failed to fetch list of interfaces from " << task.toString() << "\n";
        }

        try
        {
            q.fetchSysInfo(logger);
            found[task].have_snmpd = true;

            found[task].sysname  = q.getSysname  ();
            found[task].descr    = q.getDescr    ();
            found[task].contact  = q.getContact  ();
            found[task].location = q.getLocation ();
        } catch(const FWException &ex)
        {
            // fetch failed
            *logger << "Failed to fetch sysinfo from " << task.toString() << "\n";
        }

        try
        {
            q.fetchRoutingTable(logger);
            found[task].have_snmpd = true;
            
            vector<IPRoute> routes = q.getRoutes();
            
            qplus=0; rplus=0; dplus=0; 
            int nplus=0;

            for(vector<IPRoute>::iterator j=routes.begin(); j!=routes.end(); ++j)
            {
                IPNetwork net((*j).getDestination(), (*j).getNetmask());

                if(!special(net) && included(net.getAddress()) )
                {
                    *logger << "Network " << string(net) << " found (via " 
                            << string(((*j).isDirect())?"direct":"indirect") 
                            << " route).\n";
                    
                    if(point2point(net))
                    {
                        IPAddress c=net.getAddress();
                        if(included(c) && !alreadyseen(c) && !isvirtual(c,""))
                        {
                            if(recursive)
                            {
                                qplus++;
                                queue[c]="";
                            } else
                            {
                                rplus++;
                                found[c]=CrawlerFind();
                            }
                        } else
                        {
                            dplus++;
                        }
                    } else
                    {
                        networks.insert(net);
                        nplus++;
                    }
                }
            }

            if(qplus) *logger << "Adding "   << qplus << " hosts to queue\n"   ;
            if(nplus) *logger << "Adding "   << nplus << " networks to results\n"   ;
            if(rplus) *logger << "Adding "   << rplus << " hosts to results\n" ;
            if(dplus) *logger << "Skipping " << dplus << " hosts as duplicate, excluded or virtual\n" ;

        } catch(const FWException &ex)
        {
            // fetch failed
            *logger << "Failed to fetch routing table from " << task.toString() << "\n";
        }
    } while(recursive);

    if(skip_virtual)
        remove_virtual(logger);
    
    if(do_dns)
        bacresolve_results(logger);
    
    *logger << "SNMPCrawler - done.\n";
}

void SNMPCrawler::remove_virtual(Logger *logger) throw(FWException)
{
    *logger << "Removing virtual IPs.\n";
    for(map<IPAddress, CrawlerFind>::iterator j=found.begin(); j!=found.end(); ++j)
        if(isvirtual((*j).first, (*j).second.found_phys_addr))
            found.erase(j);
}

void SNMPCrawler::bacresolve_results(Logger *logger) throw(FWException)
{
    *logger << "Resolving found hosts.\n";
        
    set<IPAddress> resolv_set;
    for(map<IPAddress, CrawlerFind>::iterator j=found.begin(); j!=found.end(); ++j)
        resolv_set.insert((*j).first);
    DNS_bulkBackResolve_query backresq(resolv_set, dns_threads,
				       dns_retries,dns_timeout );
    resolv_set.clear();
        
    try
    {
        backresq.run_impl(logger);
    } catch(const FWException &ex)
    {
        *logger << "Resolving reported errors.\n";
    }
        
    map<IPAddress, HostEnt> resolv_res=backresq.getResult();
    for(map<IPAddress, HostEnt>::iterator j=resolv_res.begin(); j!=resolv_res.end(); ++j)
    {
        found[(*j).first].dns_ok  = true;
        found[(*j).first].dns_info=(*j).second;
    }
}


set<IPNetwork> SNMPCrawler::getNetworks()
{
    return networks;
}

map<IPAddress, CrawlerFind> SNMPCrawler::getAllIPs()
{
    return found;
}

CrawlerFind::CrawlerFind()
{
    have_snmpd = false;
    dns_ok     = false;
}

CrawlerFind::~CrawlerFind()
{
}

#else
# warning ucd-snmp library not found - SNMP functionality will be disabled.
#endif
